from rmapy.api import Client
from rmapy.folder import Folder


def create_folders(rmapy: Client, *folders):
    # Get all existing folders
    r_collection = rmapy.get_meta_items()
    r_folders = {f.ID: f for f in r_collection if isinstance(f, Folder)}

    # Create folder structure
    current_folder = None
    current_parent = ""
    for folder in folders:
        # Search for folder
        folder_found = False
        for r_id, r_folder in r_folders.items():
            if r_folder.VissibleName == folder and r_folder.Parent == current_parent:
                current_folder = r_folder
                current_parent = r_id
                folder_found = True
                break

        # Not found: create
        if not folder_found:
            current_folder = Folder(folder, Parent=current_parent)
            rmapy.create_folder(current_folder)
            current_parent = current_folder.ID

    return current_folder



import os
import shutil
import subprocess  # noqa
import tempfile
import time
from threading import Thread
from uuid import uuid4

import click
import dateparser
from rmapy.api import Client as RMClient
from rmapy.document import Document, ZipDocument
from rmapy.exceptions import AuthError
from webdav3.client import Client
from webdav3.exceptions import WebDavException

from remarkdav.config import settings
from remarkdav.crawl import CrawlRegistry, CrawlThread
from remarkdav.pdf import create_sync_status_pdf
from remarkdav.rmapy_util import create_folders
from remarkdav.sync_db import File
from remarkdav.utils import check_pdf_file, get_filename, has_extension


@click.group()
def cli():
    pass


RELEVANT_FILE_ENDINGS = ["pdf"]


class DownloadThread(Thread):
    def __init__(self, client, paths_to_sync, source):
        super().__init__()
        self.client = client
        self.paths = paths_to_sync
        self.source = source

    def run(self):
        while len(self.paths) > 0:
            path = self.paths.pop()

            filename = path["filename"]
            click.echo(f"Download {filename}")
            try:
                self.client.download_sync(
                    remote_path=path["path"], local_path=path["download_path"]
                )
            except WebDavException:
                click.secho(f"Download of {filename} failed.", fg="red")
                return
            click.echo(f"Download of {filename} finished.")

            valid_pdf = check_pdf_file(path["download_path"])
            if not valid_pdf:
                click.secho(f"File {filename} is not a valid PDF.")

            file, _ = File.get_or_create(path=path["path"], source=self.source)
            file.downloaded = True
            file.synced = False
            file.modified = path["modified"]
            file.upload_path = path["upload_path"]
            file.local_path = path["download_path"]
            file.save()


@cli.command()
def sync():
    """Run synchronisation for all mappings."""
    # Authenticate against reMarkable API
    rmapy = RMClient()

    try:
        try:
            rmapy.renew_token()
            click.secho("Successful connected to reMarkable API", fg="green")
        except AuthError:
            code = click.prompt("Please enter reMarkable one-time code")
            res = rmapy.register_device(code)
            if res:
                click.secho("Successful connected to reMarkable API", fg="green")
    except AuthError as e:
        click.secho(f"Connection error while connecting to reMarkable API: {e}", fg="red")

    mappings = settings.get("mappings", [])

    if len(mappings) < 1:
        click.secho("No valid mappings configured.", fg="red")
        return

    for mapping in mappings:
        mapping_id = mapping["id"]
        click.echo(f"Run synchronisation for config {mapping_id}")
        options = {
            "webdav_hostname": mapping["webdav"]["hostname"],
            "webdav_login": mapping["webdav"]["login"],
            "webdav_password": mapping["webdav"]["password"],
            "webdav_timeout": 60,
        }
        client = Client(options)

        # Start crawler
        registry = CrawlRegistry()
        thread = CrawlThread(client, registry, mapping["webdav"]["base_path"])
        thread.start()

        # Run crawl threads
        running = 0  # Currently running threads
        while len(registry.stack) > 0 or len(registry.threads) > 0:
            while running < 10:
                if len(registry.stack) == 0:
                    break
                thread = registry.stack.pop()
                thread.start()
            time.sleep(0.5)
            click.echo(f"Load paths: {len(registry.paths)} directories or files detected.")

        # Show count of errors
        if registry.errors > 0:
            click.secho(
                f"{registry.errors} download errors have been detected and resolved.", fg="red"
            )
        if registry.unresolved_errors > 0:
            click.secho(
                f"{registry.unresolved_errors} errors have been detected and are unresolved.",
                fg="red",
            )

        # Filter paths
        my_paths = list(
            filter(lambda path: has_extension(path["path"], RELEVANT_FILE_ENDINGS), registry.paths)
        )

        # Print found PDF paths
        print(my_paths)
        print(f"{len(my_paths)} PDF files were found.")

        # Make temp directory for downloads
        temp_directory = tempfile.mkdtemp()
        paths_to_sync = []
        for path in my_paths:
            filename = get_filename(path["path"])
            path["filename"] = filename

            # Make unique dir
            uid = str(uuid4())
            os.mkdir(os.path.join(temp_directory, uid))

            # Set download and upload paths
            path["download_path"] = os.path.join(temp_directory, uid, filename)
            path["upload_path"] = path["path"].replace(mapping["webdav"]["base_path"], "")

            sync_this_file = False

            # Check whether this file has already been synced
            qs = File.select().where(
                File.path == path["path"],
                File.source == mapping["id"],
                File.uploaded == True,  # noqa
            )
            if qs.count() >= 1:
                # Get object and modification timestamp
                file_object = qs[0]
                modified = dateparser.parse(file_object.modified)

                # If downloaded file is newer than synced file, resync this file
                if path["modified"] > modified:
                    sync_this_file = True
                    click.echo(f"Sync {filename} because its newer")
                    file_object.downloaded = False
                    file_object.uploaded = False
                    file_object.save()
            else:
                click.echo(f"Sync {filename} because it was added")
                sync_this_file = True

            if sync_this_file:
                # Add path to download list
                paths_to_sync.append(path)

        # Start download threads
        threads = []
        for i in range(10):
            thread = DownloadThread(client, paths_to_sync, mapping["id"])
            thread.start()
            threads.append(thread)

        # Wait until all files are downloaded
        for thread in threads:
            thread.join()

        # Get all files to upload
        qs = File.select().where(
            File.source == mapping["id"], File.downloaded == True, File.uploaded == False  # noqa
        )

        for file in qs:
            click.echo(f"Upload file {file.upload_path} to reMarkable Cloud")

            # Get necessary folders and the filename
            path_split = file.upload_path.split("/")
            folders = [mapping["base_folder"]] + path_split[:-1]

            create_folders(rmapy, *folders)

            # Upload new document to cloud using rmapi
            folder_path = "/" + "/".join(folders)
            subprocess.run(["rmapi", "put", os.path.abspath(file.local_path), folder_path])  # noqa

            # Save sync status
            file.uploaded = True
            file.save()

            click.secho("  Upload finished.", fg="green")

        # Create sync status
        sync_doc_name = "Sync status"
        sync_status_filename = create_sync_status_pdf(mapping)

        # Create folder
        sync_status_folder = create_folders(rmapy, mapping["base_folder"])

        # Get all existing documents
        r_collection = rmapy.get_meta_items()

        for meta_item in r_collection:
            if (
                isinstance(meta_item, Document)
                and meta_item.VissibleName == sync_doc_name
                and meta_item.Parent == sync_status_folder.ID
            ):
                # Delete old file
                rmapy.delete(meta_item)
                break

        # Upload new document to cloud
        doc = ZipDocument(doc=sync_status_filename)
        doc.metadata["VissibleName"] = sync_doc_name
        rmapy.upload(doc, sync_status_folder)

        # Clean up temp directory
        shutil.rmtree(temp_directory)


if __name__ == "__main__":
    cli()



from peewee import BooleanField, CharField, DateTimeField, Model, SqliteDatabase

from remarkdav.config import settings

db = SqliteDatabase(settings.get("db_path", "sync.db"))


class File(Model):
    path = CharField()
    source = CharField()
    modified = DateTimeField(null=True)
    downloaded = BooleanField(default=False)
    uploaded = BooleanField(default=False)
    upload_path = CharField(default="", null=True)
    local_path = CharField(default="", null=True)

    class Meta:
        database = db


db.create_tables([File], safe=True)



from threading import Thread
from uuid import uuid4

import click
import dateparser
from webdav3.exceptions import WebDavException

from remarkdav.utils import clean_path, last_part_of_path, simple_path


def parse_file_info(info: dict) -> dict:
    info["created"] = dateparser.parse(info["created"]) if info["created"] else None
    info["modified"] = dateparser.parse(info["modified"]) if info["modified"] else None

    return info


class CrawlRegistry:
    def __init__(self):
        self.paths = []
        self.threads = []
        self.stack = []
        self.errors = 0
        self.error_paths = {}
        self.unresolved_errors = 0

    def add(self, path):
        self.paths.append(path)

    def lock(self, uid: str):
        self.threads.append(uid)

    def unlock(self, uid: str):
        self.threads.remove(uid)


class CrawlThread(Thread):
    def __init__(self, client, registry: CrawlRegistry, base_path: str):
        super().__init__()
        self.base_path = base_path
        self.client = client
        self.registry = registry
        self.lock = str(uuid4())

    def run(self):
        # Add lock
        self.registry.lock(self.lock)

        # Clean base path
        base_path = clean_path(self.base_path)

        # Get sub dirs
        try:
            sub_paths = self.client.list(base_path, get_info=True)
        except WebDavException as e:
            click.secho(f"Check of {base_path} failed with {e}.", fg="red")
            self.registry.unlock(self.lock)

            if base_path in self.registry.error_paths:
                if self.registry.error_paths[base_path] >= 2:
                    click.secho(f"Check of {base_path} failed – last try.", fg="red")
                    self.registry.unresolved_errors += 1
                    return
                self.registry.error_paths[base_path] += 1
            else:
                self.registry.error_paths[base_path] = 1
                self.registry.errors += 1

            # Add new crawl thread for failed path listing
            new_thread = CrawlThread(self.client, self.registry, base_path)
            self.registry.stack.append(new_thread)
            return

        for sub_path in sub_paths:
            sub_path = parse_file_info(sub_path)

            # Skip top directory in sub dir list
            if simple_path(sub_path["path"]) == last_part_of_path(base_path):
                continue

            if sub_path not in self.registry.paths:
                # Add path to registry
                self.registry.add(sub_path)

                # Start new thread for sub dirs
                new_thread = CrawlThread(self.client, self.registry, sub_path["path"])
                self.registry.stack.append(new_thread)

        # Remove lock
        self.registry.unlock(self.lock)



import os
import re
from datetime import datetime
from tempfile import mkdtemp

from fpdf import FPDF

from remarkdav.sync_db import File


def clean_string(s: str) -> str:
    """Remove all non latin-1 characters."""
    return re.sub(r"[^\x00-\x7F\x80-\xFF\u0100-\u017F\u0180-\u024F\u1E00-\u1EFF]", "", s)


def create_sync_status_pdf(mapping: dict) -> str:
    # Get all synced filess
    qs = File.select().where(File.source == mapping["id"], File.uploaded == True)  # noqa

    # Create PDF
    pdf = FPDF()
    pdf.add_page()

    # Add title
    pdf.set_font("Arial", "B", 16)
    pdf.cell(40, 10, "Sync status", ln=1)

    # Add sync date and time
    pdf.set_font("Arial", size=12)
    pdf.cell(40, 20, f"Date and time: {datetime.now()}", ln=1)

    for file in qs:
        # Add rows with files
        pdf.cell(200, 8, clean_string(file.upload_path), ln=1)
        pdf.cell(50, 10, "", ln=0)
        pdf.cell(50, 10, str(file.modified), ln=1)

    filename = os.path.join(mkdtemp(), "status.pdf")
    pdf.output(filename, "F")

    return filename



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['remarkdav']

package_data = \
{'': ['*']}

install_requires = \
['PyPDF2>=1.26.0,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'dateparser>=1.0.0,<2.0.0',
 'dynaconf>=3.1.2,<4.0.0',
 'fpdf>=1.7.2,<2.0.0',
 'peewee>=3.14.0,<4.0.0',
 'rmapy>=0.2.2,<0.3.0',
 'webdavclient3>=3.14.5,<4.0.0']

entry_points = \
{'console_scripts': ['remarkdav = remarkdav.cli:cli']}

setup_kwargs = {
    'name': 'remarkdav',
    'version': '1.0.post0',
    'description': 'A tool to sync webdav files (only PDF) to the reMarkable cloud',
    'long_description': 'remarkdav\n=========\nThis is a small tool to sync webdav files (only PDF) to the reMarkable cloud (oneway).\n\nSetup (dev)\n-----------\n\n1. Get rmapi (Go CLI application)\n\nsee instructions at https://github.com/juruen/rmapi\n\n2. Clone from Git\n\n.. code-block::\n\n    git clone git@github.com:hansegucker/remarkdav.git\n\n3. Get poetry\n\nSee instructions at https://python-poetry.org/docs/#installation\n\n\n4. Install dependencies\n\n.. code-block::\n\n    $ cd remarkdav/ # Go to your cloned directory\n    $ poetry install\n\n5. Run\n\n.. code-block::\n\n    $ poetry run remarkdav\n\nSetup (production)\n------------------\n\n1. Get rmapi (Go CLI application)\n\nsee instructions at https://github.com/juruen/rmapi\n\n2. Get it via poetry (use pip with Python 3)\n\n.. code-block::\n\n    $ sudo pip install -G remarkdav\n\n3. Run\n\n.. code-block::\n\n    $ remarkdav\n\nConfiguration\n-------------\nCopy the example configuration file from ``settings-example.toml`` to ``/etc/remarkdav/settings.toml`` and customise it.\n\n\nCopyright\n---------\n© 2020 by Jonathan Weth <dev@jonathanweth.de>\n\nremarkdav is licenced under GPL-3.0.',
    'author': 'Jonathan Weth',
    'author_email': 'dev@jonathanweth.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List

from PyPDF2 import PdfFileReader
from PyPDF2.utils import PdfReadError


def clean_path(path: str) -> str:
    if path[-1] != "/":
        path += "/"
    return path


def simple_path(path: str) -> str:
    return path.replace("/", "")


def last_part_of_path(path: str) -> str:
    if path[-1] == "/":
        path = path[:-1]
    return path.split("/")[-1].replace("/", "")


def get_filename(path: str):
    return path.split("/")[-1]


def get_extension(path: str):
    return path.split(".")[-1]


def has_extension(filename: str, extensions: List[str]):
    return get_extension(filename) in extensions


def check_pdf_file(path: str) -> bool:
    """Check if a file is a valid PDF file."""
    try:
        PdfFileReader(open(path, "rb"))
    except PdfReadError:
        return False
    return True
